package uk.co.badgersinfoil.metaas;

/**
 * Defines the common services provided by structures which can contain
 * ActionScript 'statements'.
 */
public interface StatementContainer {
	/**
	 * Checks the syntax of the given code, and then adds the statement to
	 * the end of the current block.
	 * 
	 * @throws SyntaxException if the syntax of the given code fragment is
	 *         incorrect.
	 */
	public void addStmt(String statement);

	/**
	 * Adds a single-line comment to list of statements being generated
	 * 
	 * @param text the text of the comment (minus the initial '//') which
	 *        must not include any newline characters.
	 */
	public void addComment(String text);

	/**
	 * Adds an if-statement to the code. e.g.
	 * <pre>block.newIf("test()").addStmt("trace('success')")</pre>
	 * results in
	 * <pre>
	 * if (test()) {
	 * 	trace('success');
	 * }</pre>
	 */
	public ASIfStatement newIf(String condition);

	/**
	 * Adds a C-style for-loop to the code.  e.g.
	 * <pre>block.newFor("var i=0", "i&lt;10", "i++").addStmt("trace(i)")</pre>
	 * results in
	 * <pre>
	 * for (var i=0; i&lt;10; i++) {
	 * 	trace(i);
	 * }</pre>
	 */
	public StatementContainer newFor(String init, String condition, String update);

	/**
	 * Adds a for-in-loop to the code.  e.g.
	 * <pre>block.newForIn("var i", "myArray").addStmt("trace(i)")</pre>
	 * results in
	 * <pre>
	 * for (var i in myArray) {
	 * 	trace(i);
	 * }</pre>
	 */
	public StatementContainer newForIn(String init, String list);

	public StatementContainer newForEachIn(String init, String list);

	/**
	 * Adds a while-loop to the code.  e.g.
	 * <pre>block.newWhile("test()").addStmt("trace('hi there')")</pre>
	 * results in
	 * <pre>
	 * while (test()) {
	 * 	trace('hi there');
	 * }</pre>
	 */
	public StatementContainer newWhile(String condition);

	/**
	 * Adds a do-while-loop to the code.  e.g.
	 * <pre>block.newDoWhile("test()").addStmt("trace('hi there')")</pre>
	 * results in
	 * <pre>
	 * do {
	 * 	trace('hi there');
	 * } while (test());</pre>
	 */
	public StatementContainer newDoWhile(String condition);

	/**
	 * Adds a switch-statement to the code. See {@link ASSwitchStatement}
	 * for more information.
	 */
	public ASSwitchStatement newSwitch(String condition);

	/**
	 * Returns true if if this container currently contains at
	 * least one statement, and false if it is empty, or contains only
	 * comments and whitespace.
	 */
	public boolean containsCode();
}