package uk.co.badgersinfoil.metaas.impl;

import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.CompilationUnit;
import uk.co.badgersinfoil.metaas.SyntaxException;
import uk.co.badgersinfoil.metaas.Visibility;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListToken;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;


public class ASTFactory {

	private ASTFactory() {
		// hide default ctor
	}

	public static AS3ASTCompilationUnit synthesizeClass(String qualifiedName) {
		LinkedListTree unit = ASTUtils.newAST(AS3Parser.COMPILATION_UNIT);
		LinkedListTree pkg = ASTUtils.newAST(AS3Parser.PACKAGE, "package");
		pkg.appendToken(TokenBuilder.newSpace());
		unit.addChildWithTokens(pkg);
		String packageName = packageNameFrom(qualifiedName);
		if (packageName != null) {
			pkg.addChildWithTokens(ASTUtils.newIdent(packageName));
		}
		LinkedListTree packageBlock = newBlock();
		pkg.addChildWithTokens(packageBlock);
		String className = typeNameFrom(qualifiedName);
		
		LinkedListTree clazz = synthesizeAS3Class(className);
		ASTUtils.addChildWithIndentation(packageBlock, clazz);
		return new AS3ASTCompilationUnit(unit);
	}

	public static CompilationUnit synthesizeInterface(String qualifiedName) {
		LinkedListTree unit = ASTUtils.newAST(AS3Parser.COMPILATION_UNIT);
		LinkedListTree pkg = ASTUtils.newAST(AS3Parser.PACKAGE, "package");
		unit.addChildWithTokens(pkg);
		String packageName = packageNameFrom(qualifiedName);
		if (packageName != null) {
			pkg.addChildWithTokens(ASTUtils.newIdent(packageName));
		}
		LinkedListTree packageBlock = newBlock();
		pkg.addChildWithTokens(packageBlock);
		
		LinkedListTree clazz = ASTUtils.newAST(AS3Parser.INTERFACE_DEF);
		LinkedListTree modifiers = ASTUtils.newAST(AS3Parser.MODIFIERS);
		clazz.addChildWithTokens(modifiers);
		modifiers.addChildWithTokens(ASTUtils.newAST(AS3Parser.PUBLIC, "public"));
		clazz.addChildWithTokens(ASTUtils.newAST(AS3Parser.INTERFACE, "interface"));
		packageBlock.addChildWithTokens(clazz);
		clazz.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, typeNameFrom(qualifiedName)));
		LinkedListTree typeBlock = ASTUtils.newParentheticAST(AS3Parser.TYPE_BLOCK, AS3Parser.LCURLY, "{", AS3Parser.RCURLY, "}");
		clazz.addChildWithTokens(typeBlock);
		return new AS3ASTCompilationUnit(unit);
	}

	private static String typeNameFrom(String qualifiedName) {
		int p = qualifiedName.lastIndexOf('.');
		if (p == -1) {
			return qualifiedName;
		}
		return qualifiedName.substring(p+1);
	}

	private static LinkedListTree synthesizeAS3Class(String className) {
		LinkedListTree clazz = ASTUtils.newAST(AS3Parser.CLASS_DEF);
		LinkedListTree modifiers = ASTUtils.newAST(AS3Parser.MODIFIERS);
		clazz.addChildWithTokens(modifiers);
		LinkedListTree modPublic = ASTUtils.newAST(AS3Parser.PUBLIC, "public");
		modifiers.addChildWithTokens(modPublic);
		modifiers.appendToken(TokenBuilder.newSpace());
		clazz.appendToken(TokenBuilder.newClass());
		clazz.appendToken(TokenBuilder.newSpace());
		clazz.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, className));
		clazz.appendToken(TokenBuilder.newSpace());
		LinkedListTree typeBlock = ASTUtils.newParentheticAST(AS3Parser.TYPE_BLOCK, AS3Parser.LCURLY, "{", AS3Parser.RCURLY, "}");
		LinkedListToken typeBlockNL = TokenBuilder.newNewline();
		typeBlock.getInitialInsertionAfter().afterInsert(typeBlockNL);
		typeBlock.setInitialInsertionAfter(typeBlockNL);
		clazz.addChildWithTokens(typeBlock);
		return clazz;
	}

	private static String packageNameFrom(String qualifiedName) {
		int p = qualifiedName.lastIndexOf('.');
		if (p == -1) {
			return null;
		}
		return qualifiedName.substring(0, p);
	}

	public static ASTASMethod newClassMethod(String name, Visibility visibility, String returnType) {
		LinkedListTree def = ASTUtils.newImaginaryAST(AS3Parser.METHOD_DEF);
		def.addChildWithTokens(ASTUtils.toModifiers(visibility));
		LinkedListTree decl = ASTUtils.newAST(AS3Parser.FUNCTION, "function");
		def.addChildWithTokens(decl);
		decl.appendToken(TokenBuilder.newSpace());
		LinkedListTree methName = ASTUtils.newAST(AS3Parser.IDENT, name);
		def.addChildWithTokens(methName);
		def.addChildWithTokens(ASTUtils.newParentheticAST(AS3Parser.PARAMS, AS3Parser.LPAREN, "(", AS3Parser.RPAREN, ")"));
		if (returnType != null) {
			def.addChildWithTokens(ASTUtils.newTypeSpec(returnType));
		}
		def.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = newBlock();
		def.addChildWithTokens(block);

		return new ASTASMethod(def);
	}

	public static ASTASField newField(String name, Visibility visibility, String type) {
		if (name.indexOf('.') != -1) {
			throw new SyntaxException("field name must not contain '.'");
		}
		LinkedListTree decl = ASTUtils.newImaginaryAST(AS3Parser.VAR_DEF);
		decl.addChildWithTokens(ASTUtils.toModifiers(visibility));
		decl.addChildWithTokens(ASTUtils.newAST(AS3Parser.VAR, "var"));
		decl.appendToken(TokenBuilder.newSpace());
		LinkedListTree def = ASTUtils.newAST(AS3Parser.IDENT, name);
		decl.addChildWithTokens(def);
		if (type != null) {
			def.addChildWithTokens(ASTUtils.newTypeSpec(type));
		}
		decl.appendToken(TokenBuilder.newSemi());
		return new ASTASField(decl);
	}

	public static ASTASMethod newInterfaceMethod(String name, Visibility visibility, String returnType) {
		LinkedListTree def = ASTUtils.newImaginaryAST(AS3Parser.METHOD_DEF);
		LinkedListTree meth = ASTUtils.newAST(AS3Parser.FUNCTION, "function");
		def.addChildWithTokens(meth);
		LinkedListTree methName = ASTUtils.newAST(AS3Parser.IDENT, name);
		def.addChildWithTokens(methName);
		def.addChildWithTokens(ASTUtils.newParentheticAST(AS3Parser.PARAMS, AS3Parser.LPAREN, "(", AS3Parser.RPAREN, ")"));
		if (returnType != null) {
			def.addChildWithTokens(ASTUtils.newTypeSpec(returnType));
		}
		def.addChildWithTokens(ASTUtils.toModifiers(visibility));

		return new ASTASMethod(def);
	}

	public static LinkedListTree newBlock() {
		LinkedListTree ast = ASTUtils.newParentheticAST(AS3Parser.BLOCK,
		                                                AS3Parser.LCURLY, "{",
		                                                AS3Parser.RCURLY, "}");
		LinkedListToken nl = TokenBuilder.newNewline();
		ast.getInitialInsertionAfter().afterInsert(nl);
		ast.setInitialInsertionAfter(nl);
		return ast;
	}
}