package uk.co.badgersinfoil.metaas.impl;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import org.antlr.runtime.ANTLRReaderStream;
import org.antlr.runtime.ParserRuleReturnScope;
import org.antlr.runtime.RecognitionException;
import org.asdt.core.internal.antlr.AS3ParserLexer;
import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.Visibility;
import uk.co.badgersinfoil.metaas.SyntaxException;
import uk.co.badgersinfoil.metaas.impl.antlr.BasicListUpdateDelegate;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListToken;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTokenSource;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTokenStream;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTreeAdaptor;
import uk.co.badgersinfoil.metaas.impl.antlr.TreeTokenListUpdateDelegate;


/**
 * A collection of helper methods for dealing with AST nodes.
 */
public class ASTUtils {
	
	public static final LinkedListTreeAdaptor TREE_ADAPTOR = new LinkedListTreeAdaptor();

	static {
		TREE_ADAPTOR.setFactory(new LinkedListTreeAdaptor.Factory() {
			private BasicListUpdateDelegate basicDelegate = new BasicListUpdateDelegate();
			private ParentheticListUpdateDelegate blockDelegate = new ParentheticListUpdateDelegate(AS3Parser.LCURLY, AS3Parser.RCURLY);
			private ParentheticListUpdateDelegate paramsDelegate = new ParentheticListUpdateDelegate(AS3Parser.LPAREN, AS3Parser.RPAREN);
			public TreeTokenListUpdateDelegate create(LinkedListTree payload) {
				if (payload.isNil()) {
					return basicDelegate;
				}
				switch (payload.getType()) {
				    case AS3Parser.BLOCK:
				    case AS3Parser.TYPE_BLOCK:
					return blockDelegate;
				    case AS3Parser.PARAMS:
					return paramsDelegate;
				    default:
					return basicDelegate;
				}
			}
		});
	}

	/**
	 * Stringifies the given IDENTIFIER node.
	 */
	public static String identText(LinkedListTree ast) {
		if (ast.getType() != AS3Parser.IDENTIFIER) {
			throw new IllegalArgumentException("expected IDENTIFIER, but token was a " + tokenName(ast));
		}
		StringBuffer result = new StringBuffer();
		for (int i=0; i<ast.getChildCount(); i++) {
			LinkedListTree part = (LinkedListTree)ast.getChild(i);
			if (result.length() > 0) {
				result.append(".");
			}
			if (part.getType()==AS3Parser.DBL_COLON) {
				result.append(part.getFirstChild());
				result.append("::");
				result.append(part.getLastChild());
			} else {
				result.append(part.getText());
			}
		}
		return result.toString();
	}
	
	/**
	 * Stringifies the type name from the given TYPE_SPEC node.
	 */
	public static String typeSpecText(LinkedListTree ast) {
		if (ast.getType() != AS3Parser.TYPE_SPEC) {
			throw new IllegalArgumentException("expected TYPE_SPEC, but token was a " + tokenName(ast));
		}
		LinkedListTree type = ast.getFirstChild();
		if (type.getType() == AS3Parser.IDENTIFIER) {
			return identText(type);
		}
		// handle e.g. "void",
		return type.getText();
	}

	/**
	 * Parses the given value, returning an IDENT node.
	 * 
	 * @throws SyntaxException if the given text doesn't form a valid
	 * identifier.
	 */
	public static LinkedListTree newIdent(String value) {
		AS3Parser parser = parse(value);
		try {
			LinkedListTree result = tree(parser.identifier());
			parser.endOfFile();
			return result;
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	/**
	 * Parses the given value as an IDENT, using it to construct a new
	 * TYPE_SPEC node.
	 * 
	 * @throws SyntaxException if the given text doesn't form a valid
	 * identifier.
	 */
	public static LinkedListTree newTypeSpec(String value) {
		AS3Parser parser = parse(":" + value);
		try {
			LinkedListTree result = tree(parser.typeExpression());
			if (parser.getTokenStream().LA(1) != AS3Parser.EOF) {
				throw new SyntaxException("Unexpected tokens in type name: "+value);
			}
			return result;
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	/**
	 * Helper to access the AST produced by a parser production
	 */
	public static LinkedListTree tree(ParserRuleReturnScope returnScope) {
		return (LinkedListTree)returnScope.getTree();
	}

	/**
	 * Helper for constructing error messages
	 */
	public static String tokenName(LinkedListTree ast) {
		return tokenName(ast.getType());
	}

	/**
	 * Helper for constructing error messages
	 */
	public static String tokenName(int type) {
		return AS3Parser.tokenNames[type];
	}

	/**
	 * Returns the first child of the given AST node which has the given
	 * type, or null, if no such node exists.
	 */
	public static LinkedListTree findChildByType(LinkedListTree ast, int type) {
		return (LinkedListTree)ast.getFirstChildWithType(type);
	}

	/**
	 * Returns an ActionScript3 parser which will recognise input from the
	 * given string.
	 */
	public static AS3Parser parse(String text) {
		return parse(new StringReader(text));
	}

	/**
	 * Returns an ActionScript3 parser which will recognise input from the
	 * given reader.
	 */
	public static AS3Parser parse(Reader in) {
		ANTLRReaderStream cs;
		try {
			cs = new ANTLRReaderStream(in);
		} catch (IOException e) {
			throw new SyntaxException(e);
		}
		AS3ParserLexer lexer = new AS3ParserLexer(cs);
		LinkedListTokenSource linker = new LinkedListTokenSource(lexer);
		LinkedListTokenStream tokenStream = new LinkedListTokenStream(linker);
		AS3Parser parser = new AS3Parser(tokenStream);
		parser.setTreeAdaptor(TREE_ADAPTOR);
		return parser;
	}

	/**
	 * Constructs a new AST of the given type, initialized to contain
	 * text matching the token's name (use for non-literals only).
	 * @depricated
	 */
	public static LinkedListTree newAST(int type) {
		return newImaginaryAST(type);
	}

	/**
	 * Constructs a new AST of the given type, initialized to contain
	 * text matching the token's name (use for non-literals only).
	 * @depricated
	 */
	public static LinkedListTree newImaginaryAST(int type) {
		return (LinkedListTree)TREE_ADAPTOR.create(type, tokenName(type));
	}

	/**
	 * Constructs a new AST of the given type, initialized to contain the
	 * given text.
	 */
	public static LinkedListTree newAST(int type, String text) {
		LinkedListToken tok = TokenBuilder.newToken(type, text);
		LinkedListTree ast = (LinkedListTree)TREE_ADAPTOR.create(tok);
		return ast;
	}

	public static LinkedListTree newParentheticAST(int type,
	                                                int startType,
	                                                String startText,
	                                                int endType,
	                                                String endText) {
		LinkedListTree ast = newImaginaryAST(type);
		LinkedListToken start = TokenBuilder.newToken(startType, startText);
		ast.setStartToken(start);
		LinkedListToken stop = TokenBuilder.newToken(endType, endText);
		ast.setStopToken(stop);
		start.setNext(stop);
		ast.setInitialInsertionAfter(start);
		return ast;
	}

	/**
	 * Attempts to parse the given string as an AS3 expression.
	 * 
	 * @throws SyntaxException if the given text doesn't form a valid
	 * expression.
	 */
	public static LinkedListTree parseExpr(String expr) {
		AS3Parser parser = ASTUtils.parse(expr + ")");
		try {
			return tree(parser.expression());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	/**
	 * Attempts to parse the given string as an AS3 expression for use as
	 * a parenthesised conditional or with-statement parameter (the
	 * surronding parenthesis should not be included).
	 * 
	 * @throws SyntaxException if the given text doesn't form a valid
	 * expression.
	 */
	public static LinkedListTree parseCondition(String expr) {
		AS3Parser parser = ASTUtils.parse("(" + expr + ")");
		try {
			return tree(parser.condition());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	/**
	 * Constructs a new MODIFIERS node which represents the given
	 * visibility as an AST containing either "public", "private" or
	 * nothing.
	 */
	public static LinkedListTree toModifiers(Visibility visibility) {
		LinkedListTree modifiers = ASTUtils.newAST(AS3Parser.MODIFIERS);
		if (Visibility.PUBLIC.equals(visibility)) {
			LinkedListTree modPublic = ASTUtils.newAST(AS3Parser.PUBLIC, "public");
			modPublic.appendToken(TokenBuilder.newSpace());
			modifiers.addChildWithTokens(modPublic);
		} else if (Visibility.PRIVATE.equals(visibility)) {
			LinkedListTree modPrivate = ASTUtils.newAST(AS3Parser.PRIVATE, "private");
			modPrivate.appendToken(TokenBuilder.newSpace());
			modifiers.addChildWithTokens(modPrivate);
		}
		return modifiers;
	}

	public static void increaseIndent(LinkedListTree node, String indent) {
		LinkedListToken newStart = increaseIndentAt(node.getStartToken(), indent);
		node.setStartToken(newStart);
		increaseIndentAfterFirstLine(node, indent);
	}

	public static void increaseIndentAfterFirstLine(LinkedListTree node, String indent) {
		for (LinkedListToken tok=node.getStartToken() ; tok!=node.getStopToken(); tok=tok.getNext()) {
			switch (tok.getType()) {
			case AS3Parser.NL:
				tok = increaseIndentAt(tok.getNext(), indent);
				break;
			case AS3Parser.ML_COMMENT:
				DocCommentUtils.increaseCommentIndent(tok, indent);
				break;
			}
		}
	}

	private static LinkedListToken increaseIndentAt(LinkedListToken tok, String indentStr) {
		if (tok.getType() == AS3Parser.WS) {
			tok.setText(indentStr + tok.getText());
			return tok;
		}
		LinkedListToken indent = TokenBuilder.newWhiteSpace(indentStr);
		tok.beforeInsert(indent);
		return indent;
	}

	/**
	 * Inspects the whitespace preceeding the given tree node to try and
	 * determine its level of indentation.  Scans backwards from the
	 * startToken and returns the contents of the first whitespace token
	 * following a newline token, or the empty string if no such pattern
	 * is found.
	 */
	public static String findIndent(LinkedListTree node) {
		LinkedListToken tok = node.getStartToken();
		if (tok == null) {
			return findIndent(node.getParent());
		}
		for (; tok.getPrev()!=null; tok=tok.getPrev()) {
			if (tok.getType() == AS3Parser.NL) {
				break;
			}
		}
		if (tok.getType() == AS3Parser.WS) {
			return tok.getText();
		}
		if (tok.getType()!=AS3Parser.NL) {
			return "";
		}
		LinkedListToken startOfLine = tok.getNext();
		if (startOfLine.getType() == AS3Parser.WS) {
			return startOfLine.getText();
		}
		return "";
	}

	public static void addChildWithIndentation(LinkedListTree ast, LinkedListTree stmt) {
		LinkedListTree last = ast.getLastChild();
		String indent;
		if (last == null) {
			indent = "\t" + ASTUtils.findIndent(ast);
		} else {
			indent = ASTUtils.findIndent(last);
		}
		ASTUtils.increaseIndent(stmt, indent);
		stmt.addToken(0, TokenBuilder.newNewline());
		ast.addChildWithTokens(stmt);
	}

	public static void addChildWithIndentation(LinkedListTree ast, int index, LinkedListTree stmt) {
		LinkedListTree last = (LinkedListTree)ast.getChild(index);
		String indent;
		if (last == null) {
			indent = "\t" + ASTUtils.findIndent(ast);
		} else {
			indent = ASTUtils.findIndent(last);
		}
		ASTUtils.increaseIndent(stmt, indent);
		stmt.addToken(0, TokenBuilder.newNewline());
		ast.addChildWithTokens(index, stmt);
	}
}